import asyncio
import collections
import logging
import os
from itertools import chain

from bulk_chain.core.llm_base import BaseLM
from bulk_chain.core.service_asyncio import AsyncioService
from bulk_chain.core.service_batch import BatchIterator
from bulk_chain.core.service_data import DataService
from bulk_chain.core.service_dict import DictionaryService
from bulk_chain.core.service_json import JsonService
from bulk_chain.core.service_schema import SchemaService
from bulk_chain.core.utils import attempt_wrapper


INFER_MODES = {
    "single": lambda llm, batch, **kwargs: [llm.ask(prompt) for prompt in batch],
    "single_stream": lambda llm, batch, **kwargs: [llm.ask_stream(prompt) for prompt in batch],
    "batch": lambda llm, batch, **kwargs: llm.ask(batch),
    "batch_async": lambda llm, batch, **kwargs: AsyncioService.run_tasks(
        batch=batch, async_handler=llm.ask_async, event_loop=kwargs.get("event_loop")
    ),
    "batch_stream_async": lambda llm, batch, **kwargs: AsyncioService.run_tasks(
        batch=batch, async_handler=llm.ask_stream_async, event_loop=kwargs.get("event_loop")
    ),
}


CWD = os.getcwd()


def _iter_batch_prompts(c, batch_content_it, **kwargs):
    for ind_in_batch, entry in enumerate(batch_content_it):
        content = DataService.get_prompt_text(
            prompt=entry[c]["prompt"],
            data_dict=entry,
            handle_missed_func=kwargs["handle_missed_value_func"])
        yield ind_in_batch, content


def __handle_agen_to_gen(handle, batch, event_loop):
    """ This handler provides conversion of the async generator to generator (sync).
    """

    def __wrap_with_index(async_gens):
        async def wrapper(index, agen):
            async for item in agen:
                yield index, item
        return [wrapper(i, agen) for i, agen in enumerate(async_gens)]

    agen_list = handle(batch, event_loop=event_loop)

    it = AsyncioService.async_gen_to_iter(
        gen=AsyncioService.merge_generators(*__wrap_with_index(agen_list)),
        loop=event_loop)

    for ind_in_batch, chunk in it:
        yield ind_in_batch, str(chunk)


def __handle_gen(handle, batch, event_loop):
    """ This handler deals with the iteration of each individual element of the batch.
    """

    def _iter_entry_content(entry):
        if isinstance(entry, str):
            yield entry
        elif isinstance(entry, collections.abc.Iterable):
            for chunk in map(lambda item: str(item), entry):
                yield chunk
        else:
            raise Exception(f"Non supported type `{type(entry)}` for handling output from batch")

    for ind_in_batch, entry in enumerate(handle(batch, event_loop=event_loop)):
        for chunk in _iter_entry_content(entry=entry):
            yield ind_in_batch, chunk


def _iter_chunks(p_column, batch_content_it, **kwargs):
    handler = __handle_agen_to_gen if kwargs["infer_mode"] == "batch_stream_async" else __handle_gen
    p_batch = [item[p_column] for item in batch_content_it]
    it = handler(handle=kwargs["handle_batch_func"], batch=p_batch, event_loop=kwargs["event_loop"])
    for ind_in_batch, chunk in it:
        yield ind_in_batch, chunk


def _infer_batch(batch, batch_ind, schema, return_mode, cols=None, **kwargs):
    assert (isinstance(batch, list))

    if len(batch) == 0:
        return batch

    if cols is None:
        first_item = batch[0]
        cols = list(first_item.keys()) if cols is None else cols

    for c in cols:

        # Handling prompt column.
        if c in schema.p2r:
            content_it = _iter_batch_prompts(c=c, batch_content_it=iter(batch), **kwargs)
            for ind_in_batch, prompt in content_it:
                batch[ind_in_batch][c] = prompt

        # Handling column for inference.
        if c in schema.r2p:
            content_it = _iter_chunks(p_column=schema.r2p[c], batch_content_it=iter(batch), **kwargs)
            # Register values.
            for item in batch:
                item[c] = []
            for ind_in_batch, chunk in content_it:
                # Append batch.
                batch[ind_in_batch][c].append(chunk)
                # Returning (optional).
                if return_mode == "chunk":
                    global_ind = batch_ind * len(batch) + ind_in_batch
                    yield [global_ind, c, chunk]

            # Convert content to string.
            for item in batch:
                item[c] = "".join(item[c])

    if return_mode == "record":
        for record in batch:
            yield record

    if return_mode == "batch":
        yield batch


def iter_content(input_dicts_it, llm, schema, batch_size=1, limit_prompt=None,
                 infer_mode="batch", return_mode="batch", attempts=1, event_loop=None,
                 **kwargs):
    """ This method represent Python API aimed at application of `llm` towards
        iterator of input_dicts via cache_target that refers to the SQLite using
        the given `schema`
    """
    assert (infer_mode in INFER_MODES.keys())
    assert (return_mode in ["batch", "chunk", "record"])
    assert (isinstance(llm, BaseLM))

    # Setup event loop.
    event_loop = asyncio.get_event_loop_policy().get_event_loop() \
        if event_loop is None else event_loop

    # Quick initialization of the schema.
    if isinstance(schema, str):
        schema = JsonService.read(schema)
    if isinstance(schema, dict):
        schema = SchemaService(json_data=schema)

    prompts_it = map(
        lambda data: DictionaryService.custom_update(src_dict=dict(data), other_dict=schema.cot_args),
        input_dicts_it
    )

    handle_batch_func = lambda batch, **handle_kwargs: INFER_MODES[infer_mode](
        llm,
        DataService.limit_prompts(batch, limit=limit_prompt),
        **handle_kwargs
    )

    # Optional wrapping into attempts.
    if attempts > 1:
        # Optional setup of the logger.
        logger = logging.getLogger(__name__)
        logging.basicConfig(level=logging.INFO)

        attempt_dec = attempt_wrapper(attempts=attempts,
                                      delay_sec=kwargs.get("attempt_delay_sec", 1),
                                      logger=logger)
        handle_batch_func = attempt_dec(handle_batch_func)

    content_it = (_infer_batch(batch=batch,
                               batch_ind=batch_ind,
                               infer_mode=infer_mode,
                               handle_batch_func=handle_batch_func,
                               handle_missed_value_func=lambda *_: None,
                               return_mode=return_mode,
                               schema=schema,
                               event_loop=event_loop,
                               **kwargs)
                  for batch_ind, batch in enumerate(BatchIterator(prompts_it, batch_size=batch_size)))

    yield from chain.from_iterable(content_it)



import importlib
import logging
import sys
import time
from collections import Counter
from os.path import dirname, join, basename

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def find_by_prefix(d, key):
    """
        d: dict (str, val).
    """
    assert(isinstance(d, dict))
    assert(isinstance(key, str))

    # We first check the full match.
    for k, value in d.items():
        if k == key:
            return value

    # If we can't establish full match, then we seek by prefix.
    matches = []
    for k, value in d.items():
        if key.startswith(k):
            matches.append(k)

    if len(matches) > 1:
        raise Exception(f"There are multiple entries that are related to `{key}`: {matches}")
    if len(matches) == 0:
        raise Exception(f"No entries were found for {key}!")

    return d[matches[0]]


def iter_params(text):
    assert(isinstance(text, str))
    beg = 0
    while beg < len(text):
        try:
            pb = text.index('{', beg)
        except ValueError:
            break
        pe = text.index('}', beg+1)
        # Yield argument.
        yield text[pb+1:pe]
        beg = pe+1


def auto_import(name, is_class=False):
    """ Import from the external python packages.
    """
    def __get_module(comps_list):
        return importlib.import_module(".".join(comps_list))

    components = name.split('.')
    m = getattr(__get_module(components[:-1]), components[-1])

    return m() if is_class else m


def dynamic_init(class_filepath, class_name=None):

    # Registering path.
    target = join(dirname(class_filepath))
    logger.info(f"Adding sys path for `{target}`")
    sys.path.insert(1, target)
    class_path_list = class_filepath.split('/')

    # Composing proper class name.
    class_filename = basename(class_path_list[-1])
    if class_filename.endswith(".py"):
        class_filename = class_filename[:-len(".py")]

    # Loading library.
    class_name = class_path_list[-1].title() if class_name is None else class_name
    class_path = ".".join([class_filename, class_name])
    logger.info(f"Dynamic loading for the file and class `{class_path}`")
    cls = auto_import(class_path, is_class=False)

    return cls


def optional_limit_iter(it_data, limit=None):
    counter = Counter()
    for data in it_data:
        counter["returned"] += 1
        if limit is not None and counter["returned"] > limit:
            break
        yield data


def attempt_wrapper(attempts, delay_sec=1, logger=None):
    def decorator(func):
        def wrapper(*args, **kwargs):
            for i in range(attempts):
                try:
                    # Do action.
                    return func(*args, **kwargs)
                except Exception as e:
                    if logger is not None:
                        logger.info(f"Unable to infer the result. Try {i} out of {attempts}.")
                        logger.info(e)
                    if delay_sec is not None:
                        time.sleep(delay_sec)
            raise Exception(f"Failed after {attempts} attempts")
        return wrapper
    return decorator


class BaseLM(object):

    def __init__(self, **kwargs):
        pass

    def ask(self, content):
        """ Assumes to return str.
        """
        raise NotImplemented()

    def ask_stream(self, content):
        """ Assumes to return generator.
        """
        raise NotImplemented()

    async def ask_async(self, prompt):
        """ Assumes to return co-routine.
        """
        raise NotImplemented()

    async def ask_stream_async(self, batch):
        """ Assumes to return AsyncGenerator.
        """
        raise NotImplemented()



import unittest
from os.path import join

from bulk_chain.api import iter_content
from utils import current_dir, DEFAULT_REMOTE_LLM


class TestAPI(unittest.TestCase):


    @staticmethod
    def it_data(n):
        for i in range(n):
            yield {"ind": i, "text": "X invent sanctions against Y", "entity": "X"}

    def test_single(self):
        data_it = iter_content(input_dicts_it=self.it_data(20),
                               llm=DEFAULT_REMOTE_LLM,
                               batch_size=1,
                               infer_mode="single",
                               return_mode="batch",
                               schema=join(current_dir, "schema/default.json"))

        for data in data_it:
            print(data)

    def test_single_stream(self):
        """ Returns individual chunks.
        """
        data_it = iter_content(input_dicts_it=self.it_data(20),
                               llm=DEFAULT_REMOTE_LLM,
                               batch_size=1,
                               infer_mode="single_stream",
                               return_mode="chunk",
                               schema=join(current_dir, "schema/default.json"))

        for data in data_it:
            print(data)

    def test_batch_async(self):
        """ Return batches that passed async at the Replicate.
        """
        data_it = iter_content(input_dicts_it=self.it_data(20),
                               llm=DEFAULT_REMOTE_LLM,
                               batch_size=5,
                               infer_mode="batch_async",
                               return_mode="batch",
                               schema=join(current_dir, "schema/default.json"))

        for batch in data_it:
            for item in batch:
                print(item)

    def test_batch_stream_async(self):
        data_it = iter_content(input_dicts_it=self.it_data(20),
                               llm=DEFAULT_REMOTE_LLM,
                               batch_size=5,
                               infer_mode="batch_stream_async",
                               return_mode="chunk",
                               schema=join(current_dir, "schema/default.json"))

        for chunk_info in data_it:
            print(chunk_info)


if __name__ == '__main__':
    unittest.main()



from timeit import default_timer as timer
from utils import DEFAULT_REMOTE_LLM

start = timer()
r = ["".join([str(s) for s in DEFAULT_REMOTE_LLM.ask(f"what's the color of the {p}")])
     for p in ["sky", "ground", "water"]]
end = timer()

total = sum(len(i) for i in r)
print(f"Completed [time: {end - start}]: {len(r[0])}, {len(r[1])}, {len(r[2])}")
print(f"TPS: {total / (end-start)}")



import unittest

from bulk_chain.api import iter_content
from utils import iter_test_jsonl_samples, DEFAULT_REMOTE_LLM


class TestAPI_Streaming(unittest.TestCase):

    def test_content_iter_mode(self):

        input_dicts_it = iter_test_jsonl_samples()
        data_it = iter_content(input_dicts_it=input_dicts_it,
                               llm=DEFAULT_REMOTE_LLM,
                               batch_size=1,
                               infer_mode="single_stream",
                               return_mode="chunk",
                               attempts=2,
                               schema="schema/thor_cot_schema.json")

        for ind_in_batch, col, item in data_it:
            print("\t".join([str(ind_in_batch), str(col), item]))



import asyncio
from typing import AsyncGenerator, Any


class AsyncioService:

    @staticmethod
    async def _run_tasks_async(batch, async_handler):
        tasks = [async_handler(prompt) for prompt in batch]
        return await asyncio.gather(*tasks)

    @staticmethod
    async def _run_generator(gen, output_queue, idx):
        try:
            async for item in gen:
                await output_queue.put((idx, item))
        finally:
            await output_queue.put((idx, StopAsyncIteration))


    @staticmethod
    def run_tasks(event_loop, **tasks_kwargs):
        return event_loop.run_until_complete(AsyncioService._run_tasks_async(**tasks_kwargs))

    @staticmethod
    async def merge_generators(*gens: AsyncGenerator[Any, None]) -> AsyncGenerator[Any, None]:

        output_queue = asyncio.Queue()
        tasks = [
            asyncio.create_task(AsyncioService._run_generator(gen, output_queue, idx))
            for idx, gen in enumerate(gens)
        ]

        finished = set()
        while len(finished) < len(tasks):
            idx, item = await output_queue.get()
            if item is StopAsyncIteration:
                finished.add(idx)
            else:
                yield item

        for task in tasks:
            task.cancel()

    @staticmethod
    def async_gen_to_iter(gen, loop=None):
        """ This approach is limited. Could be considered as legacy.
            https://stackoverflow.com/questions/71580727/translating-async-generator-into-sync-one/78573267#78573267
        """

        loop_created = False
        if loop is None:
            loop_created = True
            loop = asyncio.new_event_loop()

        asyncio.set_event_loop(loop)
        try:
            while True:
                try:
                    yield loop.run_until_complete(gen.__anext__())
                except StopAsyncIteration:
                    break
        finally:
            if loop_created:
                loop.close()



class BatchIterator:

    def __init__(self, data_iter, batch_size, end_value=None, filter_func=None):
        assert (isinstance(batch_size, int) and batch_size > 0)
        assert (callable(end_value) or end_value is None)
        self.__data_iter = data_iter
        self.__index = 0
        self.__batch_size = batch_size
        self.__end_value = end_value
        self.__filter_func = (lambda _: True) if filter_func is None else filter_func

    def __iter__(self):
        return self

    def __next__(self):
        buffer = []
        while True:
            try:
                data = next(self.__data_iter)
            except StopIteration:
                break
            if self.__filter_func(data):
                buffer.append(data)
            if len(buffer) == self.__batch_size:
                break

        if len(buffer) > 0:
            self.__index += 1
            return buffer

        if self.__end_value is None:
            raise StopIteration
        else:
            return self.__end_value()



from timeit import default_timer as timer
import asyncio

from utils import DEFAULT_REMOTE_LLM


async def infer_item(prompt):
    content = []
    for chunk in DEFAULT_REMOTE_LLM.ask(prompt):
        content.append(str(chunk))
    return content


async def coro_infer_llm(prompt):
    print(f"launch: {prompt}")
    r = None
    for response in asyncio.as_completed([infer_item(prompt)]):
        r = await response
    return "".join(r)


async def main():
    batch = [f"what's the color of the {p}" for p in ["sky", "ground", "water"]]
    routines = [coro_infer_llm(p) for p in batch]
    return await asyncio.gather(*routines)


start = timer()
r = asyncio.run(main())
end = timer()

total = sum(len(i) for i in r)
print(r[0])
print(r[1])
print(r[2])
print(f"Completed [time: {end - start}]: {len(r[0])}, {len(r[1])}, {len(r[2])}")
print(f"TPS: {total / (end-start)}")
